from __future__ import annotations
import collections
import contextlib
import enum
import functools
import os
import typing
from b3.datatypes import (
    DailyBulletinType,
    DailyBulletin,
    MarketType,
    ContractCorrection,
    QuoteSize,
    Quotes
)
from b3.utils import (
    date_from_string,
    pic11v99,
    pic16v99,
    pic7v06
)

__all__ = [
    'historical_quotes_reader'
]

class _RegistryType(enum.Enum):
    HEADER  = '00'
    QUOTES  = '01'
    TRAILER = '99'

_QuotesField = collections.namedtuple('QuotesField', ['name', 'size', 'factory'])

@functools.lru_cache
def _quotes_fields():
    return (
        _QuotesField('EXCDAT', 8,  date_from_string),
        _QuotesField('CODBDI', 2,  lambda value: DailyBulletinType(int(value))),
        _QuotesField('CODNEG', 12, str),
        _QuotesField('TPMERC', 3,  MarketType),
        _QuotesField('NOMRES', 12, str),
        _QuotesField('ESPECI', 10, str),
        _QuotesField('PRAZOT', 3,  lambda value: int(value) if value != '' else None),
        _QuotesField('MODREF', 4,  str),
        _QuotesField('PREABE', 13, pic11v99),
        _QuotesField('PREMAX', 13, pic11v99),
        _QuotesField('PREMIN', 13, pic11v99),
        _QuotesField('PREMED', 13, pic11v99),
        _QuotesField('PREULT', 13, pic11v99),
        _QuotesField('PREOFC', 13, pic11v99),
        _QuotesField('PREOFV', 13, pic11v99),
        _QuotesField('TOTNEG', 5,  int),
        _QuotesField('QUATOT', 18, int),
        _QuotesField('VOLTOT', 18, pic16v99),
        _QuotesField('PREEXE', 13, pic11v99),
        _QuotesField('INDOPC', 1,  lambda value: ContractCorrection(int(value)) if int(value) != 0 else None),
        _QuotesField('DATVEN', 8,  date_from_string),
        _QuotesField('FATCOT', 7,  lambda value: QuoteSize(int(value))),
        _QuotesField('PTOEXE', 13, pic7v06),
        _QuotesField('CODISI', 12, str),
        _QuotesField('DISMES', 3,  int),
    )

def _parse_header_line(line: str) -> typing.Dict:
    return {} # Should return anything?

def _parse_quotes_line(line: str) -> typing.Dict[str, typing.Any]:
    values = {}
    pos    = 0

    for (field_name, field_size, factory) in _quotes_fields():
        stop = (pos + field_size)

        try:
            value = line[pos:stop]
            values[field_name] = factory(value.strip())
        except (IndexError, ValueError) as exc:
            print(f"failed to read field '{field_name}': {exc}")
            raise
        else:
            pos = stop

    return values

def _parse_trailer_line(line: str) -> typing.Dict:
    return {} # Should return anything?

def _parse_line(line: str) -> typing.Tuple[_RegistryType, typing.Dict[str, typing.Any]]:
    registry_type = _RegistryType(line[0:2])
    leftover_line = line[2:]
    
    if registry_type == _RegistryType.HEADER:
        return registry_type, _parse_header_line(leftover_line)

    elif registry_type == _RegistryType.QUOTES:
        return registry_type, _parse_quotes_line(leftover_line)

    elif registry_type == _RegistryType.TRAILER:
        return registry_type, _parse_trailer_line(leftover_line)

    else:
        raise ValueError(f'unknown registry type {registry_type}')

def _make_quotes(values: typing.Dict[str, typing.Any]) -> Quotes:
    return Quotes(
        open     = values['PREABE'],
        high     = values['PREMAX'],
        low      = values['PREMIN'],
        average  = values['PREMED'],
        close    = values['PREULT'],
        best_ask = values['PREOFC'],
        best_bid = values['PREOFV']
    )

def _make_daily_bulleting(values: typing.Dict[str, typing.Any]) -> DailyBulletin:
    return DailyBulletin(
        exchange_date                 = values['EXCDAT'],
        type                          = values['CODBDI'],
        isin                          = values['CODISI'],
        ticker                        = values['CODNEG'],
        market_type                   = values['TPMERC'],
        company_short_name            = values['NOMRES'],
        especification                = values['ESPECI'],
        forward_market_remaining_days = values['PRAZOT'],
        reference_currency            = values['MODREF'],
        quotes                        = _make_quotes(values),
        total_trade_market            = values['TOTNEG'],
        total_trade_count             = values['QUATOT'],
        total_trade_volume            = values['VOLTOT'],
        strike_price                  = values['PREEXE'],
        strike_price_correction_type  = values['INDOPC'],
        maturity_date                 = values['DATVEN'],
        quote_size                    = values['FATCOT'],
        strike_price_points           = values['PTOEXE'],
        distribution_number           = values['DISMES']
    )

def _reader(stream: typing.TextIO) -> typing.Generator[DailyBulletin, None, None]:
    for line in stream:
        registry_type, values = _parse_line(line)

        if registry_type == _RegistryType.QUOTES:
            yield _make_daily_bulleting(values)

@contextlib.contextmanager
def historical_quotes_reader(file: typing.Union[str, os.PathLike, typing.TextIO]):
    if isinstance(file, (str, os.PathLike)):
        try:
            file   = open(file, mode='r', encoding='ascii')
            reader = _reader(file)
            yield reader
        finally:
            file.close()
    else:
        reader = _reader(file)
        yield reader


import decimal
import enum
import dataclasses

__all__ = [
    'QuoteSize',
    'Quotes'
]

class QuoteSize(enum.IntEnum):
    UNIT     = 1
    THOUSAND = 1000

@dataclasses.dataclass(init=True)
class Quotes:
    open: decimal.Decimal
    high: decimal.Decimal
    low: decimal.Decimal
    close: decimal.Decimal
    average: decimal.Decimal
    best_ask: decimal.Decimal
    best_bid: decimal.Decimal


import base64
import datetime
import decimal

def btoa(s: str) -> bytes:
    utf8_data   = s.encode('utf-8')
    base64_data = base64.b64encode(utf8_data)

    return base64_data.decode('utf-8')

def date_from_string(date_string: str) -> datetime.date:
    return datetime.datetime.strptime(date_string, '%Y%m%d').date()

def pic_to_decimal(number: str, integral_places: int) -> decimal.Decimal:
    integral_part = number[:integral_places]
    decimal_part  = number[integral_places:]

    return decimal.Decimal(integral_part + '.' + decimal_part)

def pic11v99(number: str) -> decimal.Decimal:
    return pic_to_decimal(number, 11)

def pic16v99(number: str) -> decimal.Decimal:
    return pic_to_decimal(number, 16)

def pic7v06(number: str) -> decimal.Decimal:
    return pic_to_decimal(number, 7)


from b3.datatypes.contract_correction import *
from b3.datatypes.market              import *
from b3.datatypes.quotes              import *
from b3.datatypes.specification       import *
from b3.datatypes.daily_bulletin      import *
from b3.datatypes.company             import *


import enum

__all__ = [
    'Specification'
]

class Specification(enum.Enum):
    BDR     = 'BDR'     # Brazilian Depositary Receipt
    BNS     = 'BNS'     # Subscription bonus for miscellaneous
    BNS_BA  = 'BNS B/A' # Subscription bonus for preferred shares
    BNS_ORD = 'BNS ORD' # Subscription bonus for common shares
    ...
    ON      = 'ON'      # Nominative common shares
    ON_P    = 'ON P'    # Nominative common shares with differentiated rights
    ON_REC  = 'ON REC'  # Receipt of subscriptions for common shares
    ...


import datetime
import json
import requests
from b3.datatypes  import CompanyDetail, SecurityCode
from b3.exceptions import RequestError
from b3.utils      import btoa

__all__ = [
    'base_url',
    'company_detail'
]

def _make_service_string(cvm_code: int, language: str) -> str:
    json_obj = {
        'codeCVM': str(cvm_code),
        'language': language
    }

    json_str   = json.dumps(json_obj, indent=None, separators=(',', ':'))
    base64_str = btoa(json_str)

    return base64_str

def base_url() -> str:
    return 'https://sistemaswebb3-listados.b3.com.br/'

def company_detail(cvm_code: str) -> CompanyDetail:
    url = base_url() + 'listedCompaniesProxy/CompanyCall/GetDetail/' + _make_service_string(cvm_code, 'pt-BR')

    response = requests.get(url)
    response = response.json()

    if len(response) == 0:
        raise RequestError(f'no company found with CVM code {cvm_code}')

    security_codes = []

    try:
        for elem in response['otherCodes']:
            security_codes.append(SecurityCode(elem['code'], elem['isin']))
    except KeyError:
        pass

    try:
        return CompanyDetail(
            cnpj                    = response['cnpj'],
            cvm_code                = response['codeCVM'],
            company_name            = response['companyName'],
            company_code            = response['issuingCompany'],
            trading_name            = response['tradingName'],
            activity                = response['activity'],
            industry                = response['industryClassification'],
            market                  = response['market'],
            market_indicator        = response['marketIndicator'],
            has_bdr                 = response['hasBDR'],
            bdr_type                = response['typeBDR'],
            has_emissions           = response['hasEmissions'],
            has_quotation           = response['hasQuotation'],
            common_institution      = response['institutionCommon'],
            preferred_institution   = response['institutionPreferred'],
            status                  = response['status'],
            website                 = response['website'],
            last_date               = datetime.datetime.strptime(response['lastDate'], '%d/%m/%Y %H:%M:%S'),
            bvmf_describle_category = response['describleCategoryBVMF'],
            security_codes          = security_codes
        )
    except KeyError as exc:
        raise RequestError(f"Missing key '{exc.args[0]}'") from None


import dataclasses
import datetime
import decimal
import enum
from b3.datatypes import (
    ContractCorrection,
    Quotes,
    QuoteSize,
    Specification,
    MarketType
)

__all__ = [
    'DailyBulletinType',
    'DailyBulletin'
]


class DailyBulletinType(enum.IntEnum):
    ROUND_LOT                                          = 2  # Lote padrão
    BMFBOVESPA_REGULATIONS_SANCTION                    = 5  # Sancionadas pelos regulamentos BMF&Bovespa
    STOCKS_OF_COMPANIES_UNDER_REORGANIZATION           = 6  # Concordatárias
    EXTRAJUDICIAL_RECOVERY                             = 7  # Recuperação extrajudicial
    JUDICIAL_RECOVERY                                  = 8  # Recuperação judicial
    RAET                                               = 9  # Regime de Administração Especial Temporária
    RIGHTS_AND_RECEIPTS                                = 10 # Direitos e recibos
    INTERVENTION                                       = 11 # Intervenção
    REAL_ESTATE_FUNDS                                  = 12 # Fundos imobiliários
    _UNKNOWN13                                         = 13 #
    _REMOVED14                                         = 14 #
    BONDS                                              = 18 # Obrigações
    PRIVATE_BONUSES                                    = 22 # Bônus (privados)
    PUBLIC_BONUSES                                     = 26 # Apólices/bônus/títulos públicos
    EXERCISE_OF_INDEX_CALL_OPTIONS                     = 32 # Exercício de opções de compra de índices
    EXERCISE_OF_INDEX_PUT_OPTIONS                      = 33 # Exercício de opções de venda de índices
    EXERCISE_OF_CALL_OPTIONS                           = 38 # Exercício de opções de compra
    EXERCISE_OF_PUT_OPTIONS                            = 42 # Exercício de opções de venda
    AUCTION_OF_NONQUOTED_SECURITIES                    = 46 #
    PRIVATIZATION_AUCTION                              = 48 #
    AUCTION_OF_ECONOMICAL_RECOVERY_FUND_OF_ES_STATE    = 49 # 
    AUCTION                                            = 50 # Leilão
    FINOR_AUCTION                                      = 51 # Leilão FINOR
    FINAM_AUCTION                                      = 52 # Leilão FINAM
    FISET_AUCTION                                      = 53 # Leilão FISET
    AUCTION_OF_SHARES_IN_ARREARS                       = 54 # Leilão de ações em mora
    SALES_BY_COURT_ORDER                               = 56 # Vendas por alvará judicial
    OTHER                                              = 58 # Outros
    SHARE_SWAP                                         = 60
    GOAL                                               = 61 # Meta
    TERM                                               = 62 # Mercado a Termo
    DEBENTURES_WITH_MATURITY_DATE_OF_UP_TO_3_YEARS     = 66 # Debêntures com data de vencimento até 3 anos
    DEBENTURES_WITH_MATURITY_DATE_GREATER_THAN_3_YEARS = 68 # Debêntures com data de vencimento maior que 3 anos
    ...
    FORWARD_WITH_CONTINUOUS_MOVEMENT                   = 70  
    FORWARD_WITH_GAIN_RETENTION                        = 71 # Mercado de Futuro
    INDEX_CALL_OPTIONS                                 = 74
    INDEX_PUT_OPTIONS                                  = 75
    CALL_OPTIONS                                       = 78
    PUT_OPTIONS                                        = 82
    ODD_LOT                                            = 96

@dataclasses.dataclass(init=True, repr=True)
class DailyBulletin:
    exchange_date: datetime.date
    type: DailyBulletinType
    isin: str
    ticker: str
    market_type: MarketType
    company_short_name: str
    especification: Specification
    forward_market_remaining_days: int
    reference_currency: str
    quotes: Quotes
    total_trade_market: int
    total_trade_count: int
    total_trade_volume: decimal.Decimal
    strike_price: decimal.Decimal
    strike_price_correction_type: ContractCorrection
    maturity_date: datetime.date
    quote_size: QuoteSize
    strike_price_points: decimal.Decimal
    distribution_number: int


import dataclasses
import datetime
import typing
from b3.datatypes import SecurityCode

__all__ = [
    'CompanyDetail'
]

@dataclasses.dataclass(init=True)
class CompanyDetail:
    cnpj: str
    cvm_code: str
    company_name: str
    trading_name: str
    company_code: str
    security_codes: typing.List[SecurityCode]
    activity: str
    industry: str
    market: str
    market_indicator: str
    has_bdr: bool
    bdr_type: str
    has_emissions: bool
    has_quotation: bool
    common_institution: str
    preferred_institution: str
    status: str
    website: str
    last_date: datetime.datetime
    bvmf_describle_category: str


import dataclasses
import enum

__all__ = [
    'MarketType',
    'SecurityCode'
]

class MarketType(enum.Enum):
    CASH                             = '010'
    EXERCISE_OF_CALL_OPTIONS         = '012'
    EXERCISE_OF_PUT_OPTIONS          = '013'
    AUCTION                          = '017'
    ODD_LOT                          = '020'
    TERM                             = '030'
    FORWARD_WITH_GAIN_RETENTION      = '050'
    FORWARD_WITH_CONTINUOUS_MOVEMENT = '060'
    CALL_OPTIONS                     = '070'
    PUT_OPTIONS                      = '080'

@dataclasses.dataclass(init=True)
class SecurityCode:
    ticker: str
    isin: str